#!/usr/bin/env node

'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

require('core-js/modules/es.array.flat');
require('core-js/modules/es.array.unscopables.flat');
require('core-js/modules/es.promise');
var chalk = _interopDefault(require('chalk'));
var cosmiconfig = require('cosmiconfig');
var glob = _interopDefault(require('glob'));
var path = _interopDefault(require('path'));
var fs = require('fs-extra');
var fs__default = _interopDefault(fs);
var fs$1 = _interopDefault(require('fs'));
var resolve = _interopDefault(require('resolve'));
var Git = _interopDefault(require('simple-git/promise'));

const error = (err, code = 0) => {

  if (err instanceof Error) {
    console.error(err);
  } else {
    console.error(chalk`{red.bold ERROR:} ${err}`);
  }

  console.log("If you think this is a bug, you can report it: https://github.com/benawad/destiny/issues");
  process.exit(code);
};
const warn = msg => {
  console.log(chalk`{yellow.bold WARN:} ${msg}`);
};
const info = msg => {
  console.log(chalk`{green.bold INFO:} ${msg}`);
};
var logger = {
  error,
  warn,
  info
};

const isDirectory = filePath => fs.lstatSync(filePath).isDirectory();

const isFile = filePath => fs.lstatSync(filePath).isFile();

const globSearch = pattern => {
  const matches = glob.sync(pattern);
  const files = matches.filter(match => isFile(match));

  if (files.length === 0) {
    logger.error("Could not find any files for: " + pattern, 1);
  }

  return files;
};
/** Recursively get all file paths. */


const getFilePaths = (paths, options) => {
  let {
    detectRoots
  } = options;
  const files = [];

  while (paths.length > 0) {
    const filePath = paths.pop();
    if (filePath == null || filePath.length === 0) continue;
    const isGlobPattern = glob.hasMagic(filePath);

    if (isGlobPattern) {
      files.push(globSearch(filePath));
      continue;
    }

    if (fs.existsSync(filePath)) {
      if (isFile(filePath)) {
        files.push([filePath]);
      } else if (isDirectory(filePath)) {
        if (detectRoots) {
          const childDirectories = fs.readdirSync(path.resolve(filePath)).map(x => path.join(filePath, x)).filter(x => isDirectory(x));
          paths.push(...childDirectories);
          detectRoots = false;
        } else {
          paths.push(path.join(filePath, "/**/*.*"));
        }
      }
    } else {
      logger.error(`Unable to resolve the path: ${filePath}`);
    }
  }

  return files;
};

function findEdges(filePath) {
  const importRegex = /(?:(?:from|import)\s+["'](\.[^'"]*)["'])|(?:(?:require|import)\s*\(["'](\.[^'"]*)["']\))/gm;
  const commentRegex = /\/\*[^]*?\*\/|^.*\/\/.*$/gm;
  const edges = [];
  const fileContent = fs$1.readFileSync(filePath, {
    encoding: "utf8"
  }).toString().replace(commentRegex, "");
  let matches;

  while ((matches = importRegex.exec(fileContent)) !== null) {
    // This is necessary to avoid infinite loops with zero-width matches.
    if (matches.index === importRegex.lastIndex) {
      importRegex.lastIndex++;
    }

    edges.push([filePath, matches[1] || matches[2]]);
  }

  return edges;
}

function addEdge([start, end], graph) {
  if (!(start in graph)) {
    graph[start] = [];
  }

  graph[start].push(end);
}

const findSharedParent = paths => {
  if (paths.length === 1) return path.dirname(paths[0]);
  const fragments = paths.map(x => x.split("/"));
  const parentPaths = [];

  for (let i = 0; i < fragments[0].length; i++) {
    const fragment = fragments[0][i];

    if (fragments.every(f => f.length > i && f[i] === fragment)) {
      parentPaths.push(fragment);
    }
  }

  return parentPaths.join("/");
};

const customResolve = (id, basedir) => {
  return resolve.sync(id, {
    basedir,
    extensions: [".js", ".jsx", ".ts", ".tsx", ".svg"] // .svg is for create-react-app

  });
};

function buildGraph(files) {
  const parentFolder = findSharedParent(files);
  const graph = {};
  const oldGraph = {};
  const totalFiles = [];
  let numForwardSlashes = 0;
  let numBackSlashes = 0;

  for (let file of files) {
    if (file === ".git") {
      continue;
    }

    file = path.resolve(file);
    const start = path.relative(parentFolder, file);

    if (!(start in oldGraph)) {
      oldGraph[start] = {
        oldLocation: file,
        imports: []
      };
    }

    totalFiles.push(start);
    findEdges(file).forEach(edge => {
      if (edge[1].includes("/")) {
        numForwardSlashes++;
      } else if (edge[1].includes("\\")) {
        numBackSlashes++;
      }

      const pathWithExtension = customResolve(edge[1], path.dirname(edge[0]));
      const end = path.relative(parentFolder, pathWithExtension);
      addEdge([start, end], graph);
      oldGraph[start].imports.push({
        text: edge[1],
        resolved: end
      });
    });
  }

  return {
    parentFolder,
    graph,
    files: totalFiles,
    oldGraph,
    useForwardSlash: numForwardSlashes >= numBackSlashes ? true : false
  };
}

const isTestFile = f => /\.test\.|\.spec\.|\.stories\./.test(f);

const invertGraph = graph => {
  const invertedGraph = {};
  Object.keys(graph).forEach(k => {
    graph[k].forEach(v => {
      if (!(v in invertedGraph)) {
        invertedGraph[v] = [];
      }

      invertedGraph[v].push(k);
    });
  });
  return invertedGraph;
};

function findEntryPoints(graph) {
  const invertedGraph = invertGraph(graph);
  const possibleEntryPoints = Object.keys(graph);
  const entryPoints = possibleEntryPoints.filter(file => {
    const importedBy = invertedGraph[file];

    if (!importedBy || !importedBy.length) {
      return true;
    }

    return importedBy.every(x => isTestFile(x));
  });

  if (entryPoints.length) {
    return entryPoints;
  }

  const levelMap = {};
  possibleEntryPoints.forEach(k => {
    const n = k.split("/").length;

    if (!(n in levelMap)) {
      levelMap[n] = [];
    }

    levelMap[n].push(k);
  });

  for (let i = 1; i < 10; i++) {
    if (i in levelMap) {
      return levelMap[i];
    }
  }

  return [];
}

const moveFiles = async (newStructure, parentFolder) => {
  const git = Git(parentFolder);
  let isRepo = false;

  try {
    isRepo = await git.checkIsRepo();
  } catch {}

  for (const [k, newLocation] of Object.entries(newStructure)) {
    // skip globals
    if (k.includes("..")) {
      continue;
    }

    const oldAbsLocation = path.resolve(path.join(parentFolder, k));
    const newAbsLocation = path.resolve(path.join(parentFolder, newLocation));

    if (oldAbsLocation !== newAbsLocation) {
      // make folders
      fs__default.ensureDirSync(path.dirname(newAbsLocation));
      let shouldGitMv = false;

      if (isRepo) {
        // check if file is tracked in git
        try {
          await git.silent(true).raw(["ls-files", "--error-unmatch", oldAbsLocation]);
          shouldGitMv = true;
        } catch {}
      }

      if (shouldGitMv) {
        await git.mv(oldAbsLocation, newAbsLocation);
      } else {
        // move
        fs__default.renameSync(oldAbsLocation, newAbsLocation);
      }
    }
  }
};

const hasCycle = (node, graph, visited) => {
  const edges = graph[node];

  if (visited.has(node)) {
    return [...visited, node];
  }

  visited.add(node);
  if (edges == null || edges.length === 0) return null;

  for (const edge of edges) {
    const cycle = hasCycle(edge, graph, new Set(visited));
    if (cycle) return cycle;
  }

  return null;
};

function toFractalTree(graph, entryPoints) {
  const res = {};
  const dependencies = {};
  const testFiles = new Set();
  let containsCycle = false;

  const addDependency = (key, location) => {
    if (!Array.isArray(dependencies[key])) {
      dependencies[key] = [];
    }

    dependencies[key].push(location);
  };

  const checkDuplicates = (location, dirname, filePath) => {
    var _newLocation;

    let newLocation;

    if (Object.values(res).includes(location)) {
      newLocation = path.join(dirname, filePath.replace(/\//g, "-"));
      logger.info(`File renamed: ${filePath} -> ${newLocation}`);
    }

    return (_newLocation = newLocation) != null ? _newLocation : location;
  };

  const fn = (filePath, folderPath, graph) => {
    const basenameWithExt = path.basename(filePath);

    if (isTestFile(basenameWithExt)) {
      testFiles.add(filePath);
      return;
    }

    let folderName = path.basename(filePath, path.extname(filePath));
    const upperFolder = path.basename(path.dirname(filePath));
    const isGlobal = filePath.includes("..");
    let location = isGlobal ? filePath : path.join(folderPath, folderName === "index" && upperFolder && upperFolder !== "." ? upperFolder + path.extname(filePath) : basenameWithExt);
    location = checkDuplicates(location, folderPath, filePath);
    folderName = path.basename(location, path.extname(location));

    if (!isGlobal) {
      res[filePath] = location;
    }

    const imports = graph[filePath];

    if ((imports == null ? void 0 : imports.length) > 0) {
      const newDestination = path.join(folderPath, folderName);

      for (const importFilePath of imports) {
        if (importFilePath in res) {
          const cycle = hasCycle(importFilePath, graph, new Set());

          if (cycle) {
            containsCycle = true;
            logger.warn(`Dependency cycle detected: ${cycle.join(" -> ")}`);
          } else {
            addDependency(importFilePath, location);
          }

          continue;
        }

        addDependency(importFilePath, location);
        fn(importFilePath, newDestination, graph);
      }
    }
  };

  for (const filePath of entryPoints) {
    fn(filePath, "", graph);
  }

  if (!containsCycle) {
    Object.entries(dependencies).forEach(([k, v]) => {
      if (v.length > 1 && !k.includes("..")) {
        const parent = findSharedParent(v);
        const filename = path.basename(k);
        const upperFolder = path.basename(path.dirname(k));
        res[k] = path.join(parent, "shared", path.basename(filename, path.extname(filename)) === "index" && upperFolder && upperFolder !== "." ? upperFolder + path.extname(filename) : filename);
      }
    });
  }

  if (testFiles.size > 0) {

    for (const testFile of testFiles) {
      const [firstRelativeImport] = graph[testFile];

      if (!firstRelativeImport) {
        continue;
      }

      const testFilePath = res[firstRelativeImport];

      if (testFilePath) {
        let location = path.join(path.dirname(testFilePath), path.basename(testFile));
        location = checkDuplicates(location, path.dirname(testFilePath), testFile);
        res[testFile] = location;
      }
    }
  }

  return res;
}

/** Recursively removes all empty folders. */

function removeEmptyFolders(directory) {
  const files = fs$1.readdirSync(directory);
  if (!files) return fs$1.rmdirSync(directory);

  for (const filePath of files) {
    const fullPath = path.resolve(directory, filePath);
    const isDirectory = fs$1.lstatSync(fullPath).isDirectory();
    if (!isDirectory) continue;
    removeEmptyFolders(fullPath);
    const isEmpty = fs$1.readdirSync(fullPath).length === 0;
    if (isEmpty) fs$1.rmdirSync(fullPath);
  }
}

const makeImportPath = (p1, p2, useForwardSlashes) => {
  const fullPath = path.relative(path.dirname(p1), p2);
  const ext = path.extname(fullPath);
  let newImport = path.join(path.dirname(fullPath), path.basename(fullPath, [".js", ".jsx", ".ts", ".tsx"].includes(ext) ? ext : undefined));

  if (!newImport.startsWith(".")) {
    newImport = "./" + newImport;
  }

  if (useForwardSlashes) {
    // Replace \ with /.
    newImport = newImport.replace(/\\/g, "/");
  } else {
    // Replace / and \ with \\.
    newImport = newImport.replace(/\/|\\+/g, "\\\\");
  }

  return newImport;
};

const getNewFilePath = (file, rootOptions) => {
  for (const {
    tree,
    parentFolder
  } of rootOptions) {
    const key = path.relative(parentFolder, file);

    if (key in tree) {
      return path.resolve(path.join(parentFolder, tree[key]));
    }
  }

  return file;
};

const getNewImportPath = (absImportPath, newFilePath, rootOptions) => {
  let lastUseForwardSlash = true;

  for (const {
    tree,
    parentFolder,
    useForwardSlash
  } of rootOptions) {
    lastUseForwardSlash = useForwardSlash;
    const key = path.relative(parentFolder, absImportPath);

    if (key in tree) {
      return makeImportPath(newFilePath, path.resolve(path.join(parentFolder, tree[key])), useForwardSlash);
    }
  }

  return makeImportPath(newFilePath, absImportPath, lastUseForwardSlash);
};

const fixImports = (files, rootOptions) => {
  for (const file of files) {
    const imports = findEdges(file);

    if (!imports.length) {
      continue;
    }

    const basedir = path.dirname(file);
    const newFilePath = getNewFilePath(file, rootOptions);
    const ogText = fs.readFileSync(file).toString(); // deep copy

    let newText = ogText.repeat(1);

    for (const imp of imports) {
      const absPath = customResolve(imp[1], basedir);
      const newImportText = getNewImportPath(absPath, newFilePath, rootOptions);

      if (newImportText) {
        newText = newText.replace(`'${imp[1]}'`, `'${newImportText}'`).replace(`"${imp[1]}"`, `"${newImportText}"`);
      }
    }

    if (newText !== ogText) {
      fs.writeFileSync(file, newText);
    }
  }
};

const formatFileStructure = async (rootDirFiles, filesToEdit) => {
  const unusedFiles = [];
  const rootOptions = [];

  for (const startingFiles of rootDirFiles) {
    if (startingFiles.length <= 1) {
      continue;
    }

    logger.info("Generating a graph and fractal tree.");
    const {
      graph,
      files,
      useForwardSlash,
      parentFolder
    } = buildGraph(startingFiles);
    const tree = toFractalTree(graph, findEntryPoints(graph));
    const usedFiles = new Set(Object.entries(graph).flat(2));
    rootOptions.push({
      tree,
      useForwardSlash,
      parentFolder
    });
    files.forEach(file => {
      if (!usedFiles.has(file)) {
        unusedFiles.push(file);
      }
    });
  }

  logger.info("Fixing imports.");
  await fixImports(filesToEdit, rootOptions);

  for (const {
    tree,
    parentFolder
  } of rootOptions) {
    logger.info("Moving files.");
    await moveFiles(tree, parentFolder);
    removeEmptyFolders(parentFolder);
  }

  if (unusedFiles.length > 0) {
    logger.warn(`Found ${unusedFiles.length} unused files:` + "\n" + unusedFiles.map(file => " ".repeat(8) + file).join("\n"));
  }

  logger.info("Restructure was successful!");
};

var version = "0.2.1";

const {
  argv
} = process;
const defaultOptions = {
  detectRoots: false,
  help: false,
  version: false
};

const printVersion = () => console.log("v" + version);

const printHelp = exitCode => {
  console.log(chalk`{blue destiny} - Prettier for file structures.

{bold USAGE}

  {blue destiny} [option...] [{underline path}]

  The {underline path} argument can consist of either a {bold file path} or a {bold glob}.

{bold OPTIONS}

  -V, --version            output version number
  -h, --help               output usage information
  -dr, --detect-roots      structure after the first level
  `);
  return process.exit(exitCode);
};

const parseArgs = args => args.reduce((acc, arg) => {
  switch (arg) {
    case "-h":
    case "--help":
      acc.options.help = true;
      break;

    case "-V":
    case "--version":
      acc.options.version = true;
      break;

    case "-dr":
    case "--detect-roots":
      acc.options.detectRoots = true;
      break;

    default:
      acc.paths.push(arg);
  }

  return acc;
}, {
  options: {},
  paths: []
});

const run = async args => {
  var _ref, _cosmiconfigSync$sear;

  const config = (_ref = (_cosmiconfigSync$sear = cosmiconfig.cosmiconfigSync("destiny").search()) == null ? void 0 : _cosmiconfigSync$sear.config) != null ? _ref : {};
  const {
    options,
    paths
  } = parseArgs(args);
  const mergedOptions = { ...defaultOptions,
    ...config,
    ...options
  };
  if (mergedOptions.help) return printHelp(0);
  if (mergedOptions.version) return printVersion();
  if (paths.length === 0) return printHelp(1);
  logger.info("Resolving files.");
  const filesToRestructure = getFilePaths(paths, mergedOptions);
  const filesToEdit = filesToRestructure.flat();

  if (filesToRestructure.length === 0) {
    logger.error("Could not find any files to restructure", 1);
    return;
  }

  await formatFileStructure(filesToRestructure, filesToEdit);
};

{
  run(argv.slice(2, argv.length));
}

exports.run = run;
